# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request permissions."""

from elasticsearch_dsl import Q
from invenio_records_permissions import RecordPermissionPolicy
from invenio_records_permissions.generators import (
    AuthenticatedUser,
    Generator,
    SystemProcess,
)


class RequestCheckGenerator(Generator):
    """Generator with support for a request check function."""

    def __init__(self, check=None):
        """Constructor."""
        super().__init__()
        self._request_check_fn = check

    def _check_request(self, request):
        """Perform request check if such a function is defined."""
        if request is None:
            return False

        if self._request_check_fn is not None:
            return self._request_check_fn(request)

        return True


def _get_id(identity):
    """Get string id from identity."""
    for need in identity.provides:
        if need.method == 'id':
            return str(need.value)
    return ""


class Creator(RequestCheckGenerator):
    """Allows request makers."""

    def __init__(self, check=None, disable_query=False):
        """Constructor."""
        super().__init__(check)
        self.disable_query = disable_query

    def needs(self, request=None, **kwargs):
        """Enabling Needs."""
        if request is None:
            return []

        if self._check_request(request):
            creator = request.created_by
            if creator is not None and creator.get_need() is not None:
                return [creator.get_need()]

        return []

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as owner."""
        if self.disable_query:
            return []

        if identity:
            # We assume creators are users for now
            user_id = _get_id(identity)
            return Q("term", **{"created_by.user": user_id})
        else:
            return []


class Receiver(RequestCheckGenerator):
    """Allows request Receiver."""

    def __init__(self, check=None, disable_query=False):
        """Constructor."""
        super().__init__(check)
        self.disable_query = disable_query

    def needs(self, request=None, **kwargs):
        """Enabling Needs.

        record is a request here
        """
        if request is None:
            return []

        if self._check_request(request):
            need = request.receiver.get_need()
            if need is not None:
                return [need]

        return []

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as owner."""
        if self.disable_query:
            return []

        # It is up to community module to define community receivers
        if identity:
            user_id = _get_id(identity)
            return Q("term", **{"receiver.user": user_id})
        else:
            return []


def is_open(request):
    """Check if the request is open."""
    return request.is_open


def is_not_open(request):
    """Check if the request is closed or a draft."""
    return not request.is_open


def is_closed(request):
    """Check if the request is closed."""
    return request.is_closed


def is_not_closed(request):
    """Check if the request is open or a draft."""
    return not request.is_closed


def is_draft(request):
    """Check if the request is a draft (neither open nor closed)."""
    return not request.is_open and not request.is_closed


def is_no_draft(request):
    """Check if the request is not a draft (either open or closed)."""
    return request.is_open or request.is_closed


class Commenter(Generator):
    """Allows request event commenter."""

    def needs(self, event=None, **kwargs):
        """Enabling Needs."""
        return [event.created_by.get_need()]

        # TODO: events also need this kind of structure
        if event is None:
            return []

        creator = event.created_by
        if creator is not None and creator.get_need() is not None:
            return [creator.get_need()]

        return []

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as creator."""
        users = [n.value for n in identity.provides if n.method == "id"]
        if users:
            return Q("terms", **{"created_by.user": users})


class AllowedSearcher(Generator):
    """Any user that was allowed by the corresponding can_search.

    This is a "cheat" originally created so that Events don't need to be indexed with a
    corresponding serialized Request. The can_search permission already restricts
    who can search and in the case of Events, those can see all Events.
    """

    def query_filter(self, identity=None, **kwargs):
        """Returns all and counts on service to filter by request_id."""
        return Q("match_all")


class PermissionPolicy(RecordPermissionPolicy):
    """Permission policy."""

    # TODO:
    # - require: if receiver is community AND community restricted:
    #   community <id> member (delegate to entity?)
    can_create = [AuthenticatedUser(), SystemProcess()]
    can_read = [Creator(), Receiver(check=is_open), SystemProcess()]
    can_update = [
        Creator(check=is_not_closed),
        Receiver(check=is_open),
        SystemProcess(),
    ]
    can_delete = [Creator(check=is_draft), SystemProcess()]
    # For search, recall that _what_ identities can see is defined by `can_read`
    can_search = [AuthenticatedUser(), SystemProcess()]

    # Actions: Submit/Cancel/Accept/Decline/Expire
    can_action_submit = [Creator(check=is_draft), SystemProcess()]
    can_action_cancel = [Creator(check=is_open), SystemProcess()]
    can_action_accept = [Receiver(check=is_open), SystemProcess()]
    can_action_decline = [Receiver(check=is_open), SystemProcess()]
    can_action_expire = [SystemProcess()]

    # Request Events: Comments
    can_create_comment = [Creator(), Receiver(check=is_no_draft), SystemProcess()]
    can_update_comment = [Commenter(), SystemProcess()]
    # Might need to revisit
    can_delete_comment = [Commenter(), Receiver(), SystemProcess()]

    # Request Events: All other events
    can_create_event = [SystemProcess()]
    can_read_event = [
        # this is a search over Events and not Requests so we disable the search
        Creator(disable_query=True),
        Receiver(check=is_no_draft, disable_query=True),
        AllowedSearcher(),
        SystemProcess(),
    ]
    can_update_event = [SystemProcess()]
    can_delete_event = [SystemProcess()]
    can_search_event = [Creator(), Receiver(check=is_no_draft), SystemProcess()]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Version information for Invenio-Requests.

This file is imported by ``invenio_requests.__init__``,
and parsed by ``setup.py``.
"""

__version__ = "0.2.4"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Request type systemfield for request records."""

import inspect

from invenio_records.systemfields import SystemField

from ...customizations.base import RequestType
from ...proxies import current_registry


class RequestTypeField(SystemField):
    """Systemfield for managing the request type."""

    def __init__(self, key="type"):
        """Constructor."""
        super().__init__(key)

    def set_obj(self, instance, obj):
        """Set the request type."""
        assert isinstance(obj, RequestType)

        self.set_dictkey(instance, obj.type_id)
        self._set_cache(instance, obj)

    def __set__(self, record, value):
        """Set the request type."""
        assert record is not None

        if inspect.isclass(value):
            # if a class was passed rather than an instance, try to instantiate it
            value = value()

        if not isinstance(value, RequestType):
            raise TypeError(f"Expected 'RequestType' but got: '{type(value)}'")

        self.set_obj(record, value)

    def obj(self, instance):
        """Get the request type."""
        obj = self._get_cache(instance)
        if obj is not None:
            return obj

        type_id = self.get_dictkey(instance)
        obj = current_registry.lookup(type_id)
        self._set_cache(instance, obj)

        return obj

    def __get__(self, record, owner=None):
        """Get the request type."""
        if record is None:
            # access by class
            return self

        return self.obj(record)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request Event Schemas."""

from datetime import timezone

from invenio_records_resources.services.records.schema import BaseRecordSchema
from marshmallow import RAISE, Schema, fields, missing, validate
from marshmallow_oneofschema import OneOfSchema
from marshmallow_utils import fields as utils_fields

from ..records.api import RequestEventFormat, RequestEventType


class RequestSchema(BaseRecordSchema):
    """Schema for requests.

    Note that the payload schema and the entity reference schemas (i.e. creator,
    receiver, and topic) are dynamically constructed and injected into this schema.
    """

    # load and dump
    type = fields.String()
    title = utils_fields.SanitizedUnicode(default="")
    description = utils_fields.SanitizedUnicode()

    # Dump-only
    number = fields.String(dump_only=True)
    status = fields.String(dump_only=True)
    is_closed = fields.Boolean(dump_only=True)
    is_open = fields.Boolean(dump_only=True)
    expires_at = utils_fields.TZDateTime(
        timezone=timezone.utc, format="iso", dump_only=True
    )
    is_expired = fields.Boolean(dump_only=True)

    class Meta:
        """Schema meta."""

        unknown = RAISE


class GenericRequestSchema(RequestSchema):
    """Generic request schema.

    CAUTION: This schema should not be used for the final validation of input
    data. Use the request type's own defined schema instead.

    This schema can be used in situations where you need to do basic validation
    or dumping of a request without the payload.

    This is used e.g. in Invenio-RDM-Records for dumping a request without
    having to know the specific request type.
    """

    created_by = fields.Dict()
    receiver = fields.Dict()
    topic = fields.Dict()


class ModelFieldStr(fields.Str):
    """Marshmallow field for serializing by attribute before item."""

    def get_value(self, obj, attr, **kwargs):
        """Return obj.attr or obj[attr] in this precedence order."""
        return getattr(obj, attr, obj.get(attr, missing))


class BaseEventSchema(BaseRecordSchema):
    """Base Event schema that other schemas should inherit from."""

    type = ModelFieldStr(required=True)
    created_by = fields.Dict(dump_only=True)


class CommentSchema(BaseEventSchema):
    """Comment schema."""

    class CommentExtra(Schema):
        """Comment extras schema."""

        content = utils_fields.SanitizedHTML(
            required=True, validate=validate.Length(min=1)
        )
        format = fields.Str(
            validate=validate.OneOf(choices=[e.value for e in RequestEventFormat]),
            load_default=RequestEventFormat.HTML.value,
        )

    payload = fields.Nested(CommentExtra, required=True)


class NoExtrasSchema(BaseEventSchema):
    """No extras schema."""


class RequestEventSchema(BaseRecordSchema, OneOfSchema):
    """Schema."""

    # TODO: Make this schema hookable? how?

    type_schemas = {
        RequestEventType.COMMENT.value: CommentSchema,
        RequestEventType.ACCEPTED.value: NoExtrasSchema,
        RequestEventType.DECLINED.value: NoExtrasSchema,
        RequestEventType.CANCELLED.value: NoExtrasSchema,
        RequestEventType.REMOVED.value: NoExtrasSchema,
        RequestEventType.EXPIRED.value: NoExtrasSchema,
    }
    type_field_remove = False

    def get_obj_type(self, obj):
        """Return key of type_schemas to use given object to dump."""
        return obj.type



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

from invenio_records_resources.references.resolvers import UserResolver

from invenio_requests.services.requests import facets

from .services.permissions import PermissionPolicy

REQUESTS_PERMISSION_POLICY = PermissionPolicy
"""Override the default requests/comments permission policy."""

REQUESTS_REGISTERED_TYPES = []
"""Configuration for registered Request Types."""

REQUESTS_ENTITY_RESOLVERS = [
    UserResolver(),
]
"""Registered resolvers for resolving/creating references in request metadata."""

REQUESTS_ROUTES = {
    'details': '/requests/<pid_value>',
}
"""Invenio requests ui endpoints."""

REQUESTS_FACETS = {
    'type': {
        'facet': facets.type,
        'ui': {
            'field': 'type',
        }
    },
    'status': {
        'facet': facets.status,
        'ui': {
            'field': 'status',
        }
    },
}
"""Invenio requests facets."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfield for calculating the ``is_open`` property of a request."""

from invenio_records_resources.records.systemfields.calculated import CalculatedField

from ...customizations.base.states import RequestState


class RequestStateCalculatedField(CalculatedField):
    """Systemfield for calculating if the request is in the expected state."""

    def __init__(self, key=None, expected_state=RequestState.OPEN):
        """Constructor."""
        super().__init__(key=key, use_cache=False)
        self._expected_state = expected_state

    def calculate(self, record):
        """Check if the request's status is mapped to the expected state."""
        status = getattr(record, self.key)
        set_state = record.type.available_statuses.get(status, None)
        return set_state == self._expected_state



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests service."""

import re

from invenio_records_resources.services import RecordService, ServiceSchemaWrapper
from invenio_records_resources.services.uow import (
    RecordCommitOp,
    RecordDeleteOp,
    unit_of_work,
)

from ...customizations.base import RequestActions
from ...errors import CannotExecuteActionError
from ...proxies import current_events_service, current_registry
from ...records.api import RequestEventType
from ...resolvers.registry import ResolverRegistry
from .links import RequestLinksTemplate


class RequestsService(RecordService):
    """Requests service."""

    def check_permission(self, identity, action_name, from_action=False, **kwargs):
        """Check a permission against the identity."""
        if from_action and "request" in kwargs:
            # if we're checking permissions for a request action, we try to construct
            # the permission name from the request action's name
            type_id = kwargs["request"].type.type_id
            custom_action_name = re.sub(r"[^a-zA-Z]", "_", f"{type_id}_{action_name}")

            # use the custom action name if there's something registered
            # otherwise use the default value
            if hasattr(self.config.permission_policy_cls, custom_action_name):
                action_name = custom_action_name
            else:
                action_name = f"action_{action_name}"

        return self.permission_policy(action_name, **kwargs).allows(identity)

    @property
    def links_item_tpl(self):
        """Item links template."""
        return RequestLinksTemplate(self.config.links_item, self.config.action_link)

    @property
    def request_type_registry(self):
        """Request_type_registry."""
        return current_registry

    def _wrap_schema(self, schema):
        """Wrap schema."""
        return ServiceSchemaWrapper(self, schema)

    @unit_of_work()
    def create(
        self, identity, data, request_type, receiver, creator=None, topic=None, uow=None
    ):
        """Create a record."""
        self.require_permission(identity, "create")

        # we're not using "self.schema" b/c the schema may differ per request type!
        schema = self._wrap_schema(request_type.marshmallow_schema())
        data, errors = schema.load(
            data,
            context={"identity": identity},
            raise_errors=False,
        )

        # most of the data is initialized via the components
        request = self.record_cls.create(
            {},
            type=request_type,
        )

        creator = (
            ResolverRegistry.reference_entity(creator)
            if creator is not None
            else ResolverRegistry.reference_identity(identity)
        )

        # Run components
        self.run_components(
            "create",
            identity,
            data=data,
            record=request,
            errors=errors,
            created_by=creator,
            topic=ResolverRegistry.reference_entity(topic),
            receiver=ResolverRegistry.reference_entity(receiver),
            uow=uow,
        )

        # persist record (DB and index)
        uow.register(RecordCommitOp(request, indexer=self.indexer))

        return self.result_item(
            self,
            identity,
            request,
            schema=schema,
            links_tpl=self.links_item_tpl,
            errors=errors,
        )

    def read(self, identity, id_):
        """Retrieve a request."""
        # resolve and require permission
        request = self.record_cls.get_record(id_)
        self.require_permission(identity, "read", request=request)

        # run components
        for component in self.components:
            if hasattr(component, "read"):
                component.read(identity, record=request)

        return self.result_item(
            self,
            identity,
            request,
            schema=self._wrap_schema(request.type.marshmallow_schema()),
            links_tpl=self.links_item_tpl,
        )

    @unit_of_work()
    def update(self, identity, id_, data, revision_id=None, uow=None):
        """Update a request."""
        request = self.record_cls.get_record(id_)

        self.check_revision_id(request, revision_id)

        self.require_permission(identity, "update", request=request)

        # we're not using "self.schema" b/c the schema may differ per request type!
        schema = self._wrap_schema(request.type.marshmallow_schema())
        data, _ = schema.load(
            data,
            context={
                "identity": identity,
                "record": request,
            },
        )

        # run components
        self.run_components("update", identity, data=data, record=request, uow=uow)

        uow.register(RecordCommitOp(request, indexer=self.indexer))

        return self.result_item(
            self,
            identity,
            request,
            schema=schema,
            links_tpl=self.links_item_tpl,
        )

    @unit_of_work()
    def delete(self, identity, id_, uow=None):
        """Delete a request from database and search indexes."""
        request = self.record_cls.get_record(id_)

        # TODO do we need revisions for requests?
        # self.check_revision_id(request, revision_id)

        # check permissions
        self.require_permission(identity, "delete", request=request)

        # TODO:
        # prevent deletion if in open state?

        # run components
        self.run_components("delete", identity, record=request, uow=uow)

        uow.register(RecordDeleteOp(request, indexer=self.indexer))
        return True

    def reindex(self, identity, params=None, es_preference=None, **kwargs):
        """Reindex records matching the query parameters."""
        # TODO check later
        return super().reindex(identity, params=None, es_preference=None, **kwargs)

    def rebuild_index(self, identity):
        """Reindex all records managed by this service."""
        for req_meta in self.record_cls.model_cls.query.all():
            request = self._request_from_model(req_meta)

            if not request.is_deleted:
                self.indexer.index(request)

    @unit_of_work()
    def execute_action(self, identity, id_, action, data=None, uow=None):
        """Execute the given action for the request, if possible.

        For instance, it would be not possible to execute the specified
        action on the request, if the latter has the wrong status.
        """
        # Retrieve request and action
        request = self.record_cls.get_record(id_)
        action_obj = RequestActions.get_action(request, action)

        # check permissions
        self.require_permission(identity, action, request=request, from_action=True)

        # Check if the action *can* be executed (i.e. correct status etc.)
        if not action_obj.can_execute(identity):
            raise CannotExecuteActionError(action)

        # Create action event if defined
        # Because the action may change the request's status, this has to be done
        # before the action is executed
        event_type = action_obj.event_type
        if event_type is not None:
            current_events_service.create(
                identity, request.id, {"type": event_type}, uow=uow
            )

        # Execute action and register request for persistence.
        action_obj.execute(identity, uow)
        uow.register(RecordCommitOp(request, indexer=self.indexer))

        # Assuming that data is just for comment payload
        if data:
            comment_type = RequestEventType.COMMENT.value
            current_events_service.create(
                identity, request.id, {**data, "type": comment_type}, uow=uow
            )

        return self.result_item(
            self,
            identity,
            request,
            schema=self._wrap_schema(request.type.marshmallow_schema()),
            links_tpl=self.links_item_tpl,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

from .base import RequestAction, RequestState
from .base import RequestType as BaseRequestType
from .default import DefaultRequestType

__all__ = (
    "BaseRequestType",
    "DefaultRequestType",
    "RequestAction",
    "RequestState",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Records is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Facet definitions."""

from flask_babelex import gettext as _
from invenio_records_resources.services.records.facets import TermsFacet

type = TermsFacet(
    field='type',
    label=_('Type'),
)

status = TermsFacet(
    field='status',
    label=_('Status'),
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Component for creating request numbers."""

from invenio_records_resources.services.records.components import (
    DataComponent,
    ServiceComponent,
)


class RequestNumberComponent(ServiceComponent):
    """Component for assigning request numbers to new requests."""

    def create(self, identity, data=None, record=None, **kwargs):
        """Create identifier when record is created."""
        type(record).number.assign(record)


class EntityReferencesComponent(ServiceComponent):
    """Component for initializing a request's entity references."""

    def create(self, identity, data=None, record=None, **kwargs):
        """Initialize the entity reference fields of a request."""
        for field in ("created_by", "receiver", "topic"):
            if field in kwargs:
                setattr(record, field, kwargs[field])


class RequestDataComponent(DataComponent):
    """Request variant of DataComponent using dynamic schema."""

    def update(self, identity, data=None, record=None, **kwargs):
        """Update an existing record (request)."""
        # Clear any top-level field not set in the data.
        # Note: This ensures that if a user removes a top-level key, then we
        # also remove it from the record (since record.update() doesn't take
        # care of this). Removal of subkeys is not an issue as the
        # record.update() will update the top-level key.
        schema = record.type.marshmallow_schema()
        fields = set(schema().fields.keys())
        data_fields = set(data.keys())
        for f in fields - data_fields:
            if f in record:
                del record[f]
        # Update the remaining keys.
        record.update(data)
        # Clear None values from the record.
        record.clear_none()


class DefaultStatusComponent(ServiceComponent):
    """Component for initializing the default status of the request."""

    def create(self, identity, data=None, record=None, **kwargs):
        """Initialize the default status of the request."""
        record.status = record.type.default_status



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Enum for the various open states a request can have."""


from enum import Enum


class RequestState(Enum):
    """Enum for the various open states a request can have."""

    OPEN = "open"
    CLOSED = "closed"
    NEITHER = "neither"



#
# Copyright (C) 2021 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create tables."""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql, postgresql
from sqlalchemy_utils import JSONType, UUIDType

# revision identifiers, used by Alembic.
revision = "a14fa442680f"
down_revision = "5cd30a3503c9"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        "request_metadata",
        sa.Column(
            "created",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "json",
            sa.JSON()
            .with_variant(JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(
                    none_as_null=True, astext_type=sa.Text()), "postgresql"
            )
            .with_variant(JSONType(), "sqlite"),
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), nullable=False),
        sa.Column(
            "id", UUIDType(), nullable=False
        ),
        sa.Column("number", sa.String(length=50), nullable=True),
        sa.Column(
            "expires_at",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_request_metadata")),
    )
    op.create_index(
        op.f("ix_request_metadata_number"),
        "request_metadata",
        ["number"],
        unique=True,
    )
    op.create_table(
        "request_events",
        sa.Column(
            "created",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "id", UUIDType(), nullable=False
        ),
        sa.Column(
            "json",
            sa.JSON()
            .with_variant(JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(
                    none_as_null=True, astext_type=sa.Text()), "postgresql"
            )
            .with_variant(JSONType(), "sqlite"),
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), nullable=False),
        sa.Column("type", sa.String(length=1), nullable=False),
        sa.Column(
            "request_id", UUIDType(), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["request_id"],
            ["request_metadata.id"],
            name=op.f("fk_request_events_request_id_request_metadata"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_request_events")),
    )
    op.create_table(
        "request_number_seq",
        sa.Column(
            "value",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            autoincrement=True,
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("value", name=op.f("pk_request_number_seq")),
    )


def downgrade():
    """Downgrade database."""
    op.drop_table("request_number_seq")
    op.drop_table("request_events")
    op.drop_index(
        op.f("ix_request_metadata_number"), table_name="request_metadata"
    )
    op.drop_table("request_metadata")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

import pkg_resources

from . import config
from .registry import TypeRegistry
from .resources import (
    RequestCommentsResource,
    RequestCommentsResourceConfig,
    RequestsResource,
    RequestsResourceConfig,
)
from .services import (
    RequestEventsService,
    RequestEventsServiceConfig,
    RequestsService,
    RequestsServiceConfig,
)


class InvenioRequests:
    """Invenio-Requests extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        self.requests_service = None
        self.requests_resource = None
        self.request_comments_service = None
        self._schema_cache = {}
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        self.init_services(app)
        self.init_resources()
        self.init_registry(app)
        app.extensions["invenio-requests"] = self

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith("REQUESTS_"):
                app.config.setdefault(k, getattr(config, k))

    def service_configs(self, app):
        """Customized service configs."""
        class ServiceConfigs:
            requests = RequestsServiceConfig.build(app)
            request_events = RequestEventsServiceConfig.build(app)

        return ServiceConfigs

    def init_services(self, app):
        """Initialize the service and resource for Requests."""
        service_configs = self.service_configs(app)

        self.requests_service = RequestsService(
            config=service_configs.requests,
        )
        self.request_events_service = RequestEventsService(
            config=service_configs.request_events,
        )

    def init_resources(self):
        """Init resources."""
        self.requests_resource = RequestsResource(
            service=self.requests_service,
            config=RequestsResourceConfig,
        )

        self.request_events_resource = RequestCommentsResource(
            service=self.request_events_service,
            config=RequestCommentsResourceConfig,
        )

    def init_registry(self, app):
        """Initialize the registry for Requests per type."""
        self.request_type_registry = TypeRegistry(
            app.config["REQUESTS_REGISTERED_TYPES"]
        )
        self.entity_resolvers_registry = TypeRegistry(
            app.config["REQUESTS_ENTITY_RESOLVERS"]
        )
        # Load from entry points
        register_entry_point(self.request_type_registry, "invenio_requests.types")
        register_entry_point(
            self.entity_resolvers_registry, "invenio_requests.entity_resolvers"
        )


def register_entry_point(registry, ep_name):
    """Register types from an entry point."""
    for ep in pkg_resources.iter_entry_points(ep_name):
        type_cls = ep.load()
        registry.register_type(type_cls())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""RequestActions define code to be executed when performing actions on requests."""


from invenio_access.permissions import system_process

from ...records.api import RequestEventType
from ..base import RequestAction


class SubmitAction(RequestAction):
    """Submit a request."""

    status_from = ['draft']
    status_to = 'open'


class AcceptAction(RequestAction):
    """Decline a request."""

    status_from = ['open']
    status_to = 'accepted'
    event_type = RequestEventType.ACCEPTED.value


class DeclineAction(RequestAction):
    """Decline a request."""

    status_from = ['open']
    status_to = 'declined'
    event_type = RequestEventType.DECLINED.value


class CancelAction(RequestAction):
    """Cancel a request."""

    status_from = ['open']
    status_to = 'cancelled'
    event_type = RequestEventType.CANCELLED.value


class ExpireAction(RequestAction):
    """Expire a request."""

    status_from = ['open']
    status_to = 'expired'
    event_type = RequestEventType.EXPIRED.value

    def can_execute(self, identity):
        """Check whether the action can be executed."""
        is_system_process = system_process in identity.provides
        return super().can_execute(identity) and is_system_process



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

from .actions import (
    AcceptAction,
    CancelAction,
    DeclineAction,
    ExpireAction,
    SubmitAction,
)
from .request_types import DefaultRequestType

__all__ = (
    "DefaultRequestType",
    "AcceptAction",
    "CancelAction",
    "DeclineAction",
    "ExpireAction",
    "SubmitAction",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Base class for customizable actions on requests."""

from ...errors import NoSuchActionError


class RequestAction:
    """Base class for actions on requests."""

    status_from = None
    """Required status of a request to run this action."""

    status_to = 'draft'
    """Status after execution of the action."""

    event_type = None
    """Defines an event type which will be logged if defined."""

    def __init__(self, request):
        """Constructor."""
        self.request = request

    def can_execute(self, identity):
        """Check whether the action can be executed.

        This is mostly intended to be a hook for checking prerequisites
        (think, for instance, CI/CD pipeline runs).
        :param identity: The identity of the executor.
        :return: True if the action can be executed, False otherwise.
        """
        return self.status_from is None or \
            self.request.status in self.status_from

    def execute(self, identity, uow):
        """Execute the request action.

        :param identity: The identity of the executor.
        :param data: The passed input to the action.
        """
        self.request.status = self.status_to


class RequestActions:
    """Namespace for RequestActions static calls."""

    @classmethod
    def get_action(cls, request, action_name):
        """Get the action registered under the given name.

        :param action_name: The registered name of the action.
        :return: The action registered under the given name.
        """
        try:
            return request.type.available_actions[action_name](request)
        except KeyError:
            raise NoSuchActionError(action=action_name)

    @classmethod
    def can_execute(cls, identity, request, action_name):
        """Check wether identity and request can execute action.

        Perhaps data is sometimes useful for that check, so also included.
        """
        return cls.get_action(request, action_name).can_execute(identity)

    @classmethod
    def execute(cls, identity, request, action_name, uow):
        """Have identity execute on request the action with the data."""
        return cls.get_action(request, action_name).execute(identity, uow)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base class for creating custom types of requests.

The `RequestType` classes are the most important part in the customization/extension
mechanism for custom types of requests.
TODO explain what can be done here, and how!
"""


from ..base import RequestState, RequestType
from .actions import (
    AcceptAction,
    CancelAction,
    DeclineAction,
    ExpireAction,
    SubmitAction,
)


class DefaultRequestType(RequestType):
    """Default class for request types."""

    type_id = "default-request"
    """The unique and constant identifier for this type of requests.

    Since this property is used to map generic chunks of data from the database
    (i.e. the request model entries) to their correct `RequestType`, this should
    be a globally unique value.
    By convention, this would be the name of the package in which the custom
    `RequestType` is defined as prefix, together with a suffix related to the
    `RequestType`.
    Further, it should be constant after the first release of the package
    (otherwise, requests created with the old value will no longer be able to be
    mapped to their `RequestType`).
    """

    name = "Default Request"
    """The human-readable name for this type of requests."""

    available_statuses = {
        "draft": RequestState.NEITHER,
        "open": RequestState.OPEN,
        "cancelled": RequestState.CLOSED,
        "declined": RequestState.CLOSED,
        "accepted": RequestState.CLOSED,
        "expired": RequestState.CLOSED,
    }
    """Available statuses for the Request.

    The keys in this dictionary is the set of available statuses, and their
    values are indicators whether this Request is still considered to be
    "open" in this state.
    """

    default_status = "draft"
    """The default status for new requests of this type.

    This must be set to one of the available statuses for the custom request type.
    """

    available_actions = {
        "submit": SubmitAction,
        "accept": AcceptAction,
        "cancel": CancelAction,
        "decline": DeclineAction,
        "expire": ExpireAction,
    }
    """Available actions for this Request.

    The keys are the internal identifiers for the actions, the values are
    the actual RequestAction classes (not objects).
    Whenever an action is looked up, a new object of the registered
    RequestAction class is instantiated with the current Request object as
    argument.
    """



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Identity systemfield."""

from invenio_records.systemfields import ModelField


class IdentityField(ModelField):
    """Systemfield for managing the request's number."""

    def assign(self, record, **kwargs):
        """Generate and assign a new identifier if none is set yet."""
        try:
            value = getattr(record.model, self.model_field_name)
        except AttributeError:
            value = None

        if value is None:
            value = record.type.generate_request_number(record, **kwargs)
            self._set(record.model, value)

        return value



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

import os

from setuptools import find_packages, setup

readme = open("README.rst").read()
history = open("CHANGES.rst").read()

tests_require = [
    "black>=20.8b1",
    "invenio-app>=1.3.2",
    "pytest-invenio>=1.4.2",
]

records_resources_version = ">=0.18.3,<0.19"

extras_require = {
    "docs": [
        "Sphinx>=4.2.0,<5.0",
    ],
    "elasticsearch6": [
        f"invenio-records-resources[elasticsearch6]{records_resources_version}"
    ],
    "elasticsearch7": [
        f"invenio-records-resources[elasticsearch7]{records_resources_version}"
    ],
    "mysql": [f"invenio-records-resources[mysql]{records_resources_version}"],
    "postgresql": [f"invenio-records-resources[postgresql]{records_resources_version}"],
    "sqlite": [f"invenio-records-resources[sqlite]{records_resources_version}"],
    "tests": tests_require,
}

extras_require["all"] = []
for name, reqs in extras_require.items():
    if name in (
        "elasticsearch6",
        "elasticsearch7",
        "mysql",
        "postgresql",
        "sqlite",
    ):
        continue
    extras_require["all"].extend(reqs)

setup_requires = [
    "Babel>=2.8",
]

install_requires = [
    "invenio-db>=1.0.11,<2.0",
    f"invenio-records-resources{records_resources_version}",
]

packages = find_packages()


# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join("invenio_requests", "version.py"), "rt") as fp:
    exec(fp.read(), g)
    version = g["__version__"]

setup(
    name="invenio-requests",
    version=version,
    description=__doc__,
    long_description=readme + "\n\n" + history,
    keywords="invenio requests",
    license="MIT",
    author="CERN",
    author_email="info@inveniosoftware.org",
    url="https://github.com/inveniosoftware/invenio-requests",
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    entry_points={
        "invenio_base.apps": [
            "invenio_requests = invenio_requests:InvenioRequests",
        ],
        "invenio_base.api_apps": [
            "invenio_requests = invenio_requests:InvenioRequests",
        ],
        "invenio_base.api_blueprints": [
            "invenio_requests = invenio_requests.views:create_requests_bp",
            "invenio_request_events = invenio_requests.views:create_request_events_bp"  # noqa
        ],
        "invenio_db.alembic": [
            "invenio_requests = invenio_requests:alembic",
        ],
        "invenio_db.models": [
            "invenio_requests = invenio_requests.records.models",
        ],
        "invenio_jsonschemas.schemas": [
            "jsonschemas = invenio_requests.records.jsonschemas",
        ],
        "invenio_search.mappings": [
            "requests = invenio_requests.records.mappings",
            "request_events = invenio_requests.records.mappings",
        ],
        "invenio_i18n.translations": [
            "messages = invenio_requests",
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 1 - Planning",
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""RequestEvents Service."""

from invenio_access.permissions import system_process
from invenio_records_resources.services import RecordService
from invenio_records_resources.services.base.links import LinksTemplate
from invenio_records_resources.services.uow import (
    RecordCommitOp,
    RecordDeleteOp,
    unit_of_work,
)

from ...records.api import RequestEventType


class RequestEventsService(RecordService):
    """Request Events service."""

    @unit_of_work()
    def create(self, identity, request_id, data, uow=None):
        """Create a request event.

        :param request_id: Identifier of the request (data-layer id).
        :param identity: Identity of user creating the event.
        :param dict data: Input data according to the data schema.
        """
        request = self._get_request(request_id)
        permission = self._get_permission("create", data["type"])
        self.require_permission(identity, permission, request=request)

        # Validate data (if there are errors, .load() raises)
        data, errors = self.schema.load(
            data,
            context={"identity": identity},
        )

        # It's the components that save the actual data in the record.
        record = self.record_cls.create(
            {},
            request=request.model,
            request_id=str(request_id),
            type=data["type"],
        )

        creator = self._get_creator(identity)

        # Run components
        self.run_components(
            "create",
            identity=identity,
            record=record,
            request=request,
            data=data,
            created_by=creator,
            uow=uow,
        )

        # Persist record (DB and index)
        uow.register(RecordCommitOp(record, indexer=self.indexer))

        return self.result_item(
            self,
            identity,
            record,
            links_tpl=self.links_item_tpl,
        )

    def read(self, identity, id_):
        """Retrieve a record."""
        record = self._get_event(id_)
        request = self._get_request(record.request_id)

        # Same "read_event" permission for all types of events
        self.require_permission(identity, "read_event", request=request)

        return self.result_item(
            self,
            identity,
            record,
            links_tpl=self.links_item_tpl,
        )

    @unit_of_work()
    def update(self, identity, id_, data, revision_id=None, uow=None):
        """Update an event (except for type)."""
        record = self._get_event(id_)
        request = self._get_request(record.request.id)
        data["type"] = record.type  # this service method doesn't allow type change

        self.check_revision_id(record, revision_id)

        # Permissions
        permission = self._get_permission("update", record.type)
        self.require_permission(identity, permission, event=record)

        data, _ = self.schema.load(
            data,
            context=dict(
                identity=identity,
                record=record,
            ),
        )

        # Run components
        self.run_components(
            "update",
            identity=identity,
            record=record,
            data=data,
            uow=uow,
        )

        uow.register(RecordCommitOp(record, indexer=self.indexer))

        return self.result_item(
            self,
            identity,
            record,
            links_tpl=self.links_item_tpl,
        )

    @unit_of_work()
    def delete(self, identity, id_, revision_id=None, uow=None):
        """Delete an event from database and search indexes.

        Deleting a comment is wiping the content and marking it deleted.
        Deleting an accepted/cancelled/declined is just wiping the content.
        Deleting another event is really deleting them.

        We may want to add a parameter to assert a kind of event is deleted
        to prevent the weird semantic of using the comments REST API to
        delete an event (which is only possible for an admin anyway).
        """
        record = self._get_event(id_)
        request = self._get_request(record.request_id)

        self.check_revision_id(record, revision_id)

        # Permissions
        permission = self._get_permission("delete", record.type)
        self.require_permission(identity, permission, request=request, event=record)

        if record.type == RequestEventType.COMMENT.value:
            record["payload"]["content"] = ""
            record.type = RequestEventType.REMOVED.value
            uow.register(
                RecordCommitOp(record, indexer=self.indexer, index_refresh=True)
            )
        else:
            uow.register(
                RecordDeleteOp(
                    record, force=True, indexer=self.indexer, index_refresh=True
                )
            )

        # Even though we don't always completely remove the RequestEvent
        # we return as though we did.
        return True

    def search(self, identity, request_id, params=None, es_preference=None, **kwargs):
        """Search for events (optionally of request_id) matching the querystring."""
        params = params or {}
        params.setdefault("sort", "oldest")

        # Permissions
        request = self._get_request(request_id) if request_id else None
        self.require_permission(identity, "search_event", request=request)

        # Prepare and execute the search
        search = self._search(
            "search",
            identity,
            params,
            es_preference,
            permission_action="read_event",
            **kwargs,
        )
        if request_id:
            search = search.filter("term", request_id=str(request_id))
        search_result = search.execute()

        return self.result_list(
            self,
            identity,
            search_result,
            params,
            links_tpl=LinksTemplate(
                self.config.links_search,
                context={"request_id": request_id, "args": params},
            ),
            links_item_tpl=self.links_item_tpl,
        )

    # Utilities
    @property
    def request_cls(self):
        """Get associated request class."""
        return self.config.request_cls

    def _get_permission(self, action, event_type):
        """Get associated permission.

        Needed to distinguish between kinds of events.
        """
        if (
            (event_type == RequestEventType.COMMENT.value)
            and (action in ["create", "update", "delete"])
        ):
            return f"{action}_comment"

        if action == "create":
            if event_type == RequestEventType.ACCEPTED.value:
                return "action_accept"
            elif event_type == RequestEventType.DECLINED.value:
                return "action_decline"
            elif event_type == RequestEventType.CANCELLED.value:
                return "action_cancel"

        return f"{action}_event"

    def _get_request(self, request_id):
        """Get associated request."""
        return self.request_cls.get_record(request_id)

    def _get_event(self, event_id, with_deleted=True):
        """Get associated event_id."""
        return self.record_cls.get_record(event_id, with_deleted=with_deleted)

    def _get_creator(self, identity):
        """Get the creator dict from the identity."""
        if system_process in identity.provides:
            return None  # TODO: Change this for some agreed value
        else:
            return {"user": str(identity.id)}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Base classes for request customization."""

from .actions import RequestAction, RequestActions
from .request_types import RequestType
from .states import RequestState

__all__ = (
    "RequestAction",
    "RequestActions",
    "RequestType",
    "RequestState",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request Events Service Config."""

from invenio_records_resources.services import Link, RecordServiceConfig
from invenio_records_resources.services.records.components import DataComponent
from invenio_records_resources.services.records.links import pagination_links
from invenio_records_resources.services.records.results import RecordItem

from ...records.api import Request, RequestEvent
from ..configurator import ConfiguratorMixin, FromConfig
from ..permissions import PermissionPolicy
from ..requests.components import EntityReferencesComponent
from ..schemas import RequestEventSchema


class RequestEventItem(RecordItem):
    """RequestEvent result item."""

    @property
    def id(self):
        """Id property."""
        return self._record.id


class RequestEventLink(Link):
    """Link variables setter for RequestEvent links."""

    @staticmethod
    def vars(record, vars):
        """Variables for the URI template."""
        vars.update({"id": record.id, "request_id": record.request_id})


class RequestEventsServiceConfig(RecordServiceConfig, ConfiguratorMixin):
    """Config."""

    request_cls = Request
    permission_policy_cls = FromConfig(
        "REQUESTS_PERMISSION_POLICY", default=PermissionPolicy
    )
    schema = RequestEventSchema
    record_cls = RequestEvent
    components = [
        DataComponent,
        EntityReferencesComponent,  # only used for created_by
    ]
    result_item_cls = RequestEventItem

    # ResultItem configurations
    links_item = {
        "self": RequestEventLink("{+api}/requests/{request_id}/comments/{id}"),
    }
    links_search = pagination_links("{+api}/requests/{request_id}/timeline{?args*}")
